from datetime import datetime

from django import test
from django.contrib.contenttypes.models import ContentType
from django.test.utils import override_settings

from .models import Poll, Choice

from events_watcher.utils import load_class


class DatabaseEventTest(test.TestCase):
    def setUp(self):
        self.backend = load_class('events_watcher.backends.database.DatabaseBackend')()
        self.backend.purge()

    def test_create_basic_event(self):
        poll = Poll.objects.create(question='WHAT?', pub_date=datetime.now())

        self.assertEqual(poll.question, 'WHAT?')

        event = self.backend.add('first_support', poll)

        self.assertEqual(event.content_object, poll)

        self.assertEqual(event.object_id, poll.pk)

        self.assertEqual(event.content_type, ContentType.objects.get_for_model(poll))

        # check if the date is not None even if we don't set it in add()
        self.assertEqual(event.date is None, False)

        current_date = datetime.now()
        event = self.backend.add('last_support', poll, date=current_date)

        self.assertEqual(event.date, current_date)

    def test_list_event(self):
        poll = Poll.objects.create(question='WHAT?', pub_date=datetime.now())

        event = self.backend.add('first_support', poll)

        self.assertEqual(self.backend.retrieve('first_support', poll), event)

        event_list = self.backend.list(poll)

        self.assertEqual(list(event_list), [event])

        event_list = self.backend.list('first_support')

        self.assertEqual(list(event_list), [event])

        polls = []
        event_list = []

        questions = (
            ('ALLO?', 'last_support'),
            ('HERE?', 'first_fan'),
            ('WHERE?', 'last_fan'),
            ('DONDE?', 'last_fan'),
            ('AQUI?', 'last_fan'),
        )

        for question, name in questions:
            poll = Poll.objects.create(question=question, pub_date=datetime.now())
            polls.append(poll)
            event_list.append(self.backend.add(name, poll))

        self.assertEqual(list(self.backend.list('last_fan')), [event_list[2], event_list[3], event_list[4]])

        self.assertEqual(list(self.backend.list(polls[2])), [event_list[2]])

        self.assertEqual(list(self.backend.list(Poll)), [event] + event_list)

        self.assertEqual(list(self.backend.list(5457445)), [])

    def test_remove_event(self):
        poll = Poll.objects.create(question='WHAT?', pub_date=datetime.now())
        self.assertEqual(poll.question, 'WHAT?')

        self.backend.add('first_support', poll)

        self.backend.remove('first_support')

        list_obj = self.backend.list('first_support')

        self.assertEqual(list(list_obj), [])

        poll4 = Poll.objects.create(question='WHERE?', pub_date=datetime.now())
        poll5 = Poll.objects.create(question='DONDE?', pub_date=datetime.now())
        event4 = self.backend.add('last_fan', poll4)
        event5 = self.backend.add('last_fan', poll5)

        list_obj = self.backend.list('last_fan')

        self.assertEqual(list(list_obj), [event4, event5])

        self.backend.remove('last_fan')

        list_obj_2 = self.backend.list('last_fan')
        self.assertEqual(list(list_obj_2), [])

    def test_watch(self):
        poll = Poll.objects.create(question='Y U NO WORK?', pub_date=datetime.now())

        choice = Choice.objects.create(poll=poll, choice='YES', votes=0)

        self.backend.watch(Choice,
                           'votes_higher_than_zero',
                           lambda initial_data, instance, created: created or initial_data['votes'] <= 0 and instance.votes > 0)

        choice.votes = 1
        choice.save()

        events = self.backend.list('votes_higher_than_zero')

        self.assertEqual(len(events), 1)

        event = events[0]

        self.assertEqual(event.name, 'votes_higher_than_zero')
        self.assertEqual(event.content_object, choice)

        choice.save()

        events = self.backend.list('votes_higher_than_zero')

        self.assertEqual(len(events), 1)

        choice = Choice.objects.create(poll=poll, choice='YES', votes=1)

        events = self.backend.list('votes_higher_than_zero')

        self.assertEqual(len(events), 2)

    def test_watch_with_callback(self):
        self.initial = False

        def callback(event, **kwargs):
            self.initial = True

        self.backend.watch(Choice,
                           'votes_higher_than_zero',
                           lambda initial_data, instance, created: created or initial_data['votes'] <= 0 and instance.votes > 0,
                           callback=callback)

        poll = Poll.objects.create(question='Y U NO WORK?', pub_date=datetime.now())

        choice = Choice.objects.create(poll=poll, choice='YES', votes=0)
        choice.votes = 1
        choice.save()

        self.assertTrue(self.initial)


@override_settings(INSTALLED_APPS=[],
                   events_watcher_BACKEND='events_watcher.backend.backends.redis.RedisBackend')
class RedisEventTest(DatabaseEventTest):
    def setUp(self):
        self.backend = load_class('events_watcher.backends.redis.RedisBackend')()
        import redisco
        redisco.connection.flushdb()

        self.backend.purge()



import inspect

from django.core.exceptions import ImproperlyConfigured
from django.contrib.contenttypes.models import ContentType
from django.db.models import Model

from ..base import Backend

from events_watcher import settings
from events_watcher.utils import load_class

from datetime import datetime


class DatabaseBackend(Backend):
    filters = [
        (lambda instance: isinstance(instance, str), 'list_from_string'),
        (lambda instance: isinstance(instance, Model), 'list_from_instance'),
        (lambda instance: inspect.isclass(instance) and issubclass(instance, Model), 'list_from_class'),
    ]

    def __init__(self):
        super(DatabaseBackend, self).__init__()

        self.model = load_class(settings.DATABASE_CONNECTION_CLASS)

        if not self.model._meta.installed:
            raise ImproperlyConfigured(
                "The events_watcher.backends.database app isn't installed "
                "correctly. Make sure it's in your INSTALLED_APPS settings.")

    def add(self, name, instance, date=None):
        if not date:
            date = datetime.now()

        content_type = ContentType.objects.get_for_model(instance)

        event = self.model.objects.create(object_id=instance.pk,
                                          name=name,
                                          date=date,
                                          content_type=content_type)

        return event

    def list_from_string(self, string):
        return self.model.objects.filter(name=string)

    def list_from_instance(self, instance):
        return self.model.objects.filter(content_type=ContentType.objects.get_for_model(instance),
                                         object_id=instance.pk)

    def list_from_class(self, klass):
        return self.model.objects.filter(content_type=ContentType.objects.get_for_model(klass))

    def remove(self, instance):
        return self.list(instance).delete()

    def retrieve(self, name, instance):
        try:
            return self.model.objects.get(content_type=ContentType.objects.get_for_model(instance),
                                          object_id=instance.pk, name=name)
        except self.model.DoesNotExist:
            return False



from collections import defaultdict

from django.db.models.signals import post_save, pre_save


class Backend(object):
    filters = []

    def __init__(self):
        self.purge()

    def purge(self):
        self._callbacks = defaultdict(list)
        self._initial_data = {}

    def add(self, key, instance, date=None):
        """
        Add new event to the backend store and return the value.
        """
        raise NotImplementedError

    def list(self, instance):
        """
        Retrieve a list of events given an instance
        """
        for condition, method in self.filters:
            if condition(instance) and hasattr(self, method):
                return getattr(self, method)(instance)

        return []

    def remove(self, instance):
        """
        Remove all events from the backend store givent an instance
        """
        raise NotImplementedError

    def retrieve(self, name, instance):
        """
        Retrieve events from the backend store given a name and an instance.
        """
        raise NotImplementedError

    def make_key_from_class(self, klass):
        """
        Make a formatted key from a class
        """
        return u'%s.%s' % (klass._meta.app_label,
                           klass.__name__)

    def make_key_from_instance(self, instance):
        """
        Make a formatted key from an instance
        """
        return u'%s.%s' % (instance._meta.app_label,
                           instance.__class__.__name__)

    def make_key_id_from_instance(self, instance):
        """
        Make a formatted key with an id from an instance
        """
        return u'%s:%s' % (self.make_key_from_instance(instance),
                           instance.pk)

    def watch(self, klass, key, condition, callback=None):
        pre_save.connect(self._on_pre_save, sender=klass)
        post_save.connect(self._on_post_save, sender=klass)

        self._callbacks[klass].append((key, condition, callback))

    def _on_post_save(self, sender, instance, **kwargs):
        klass = instance.__class__
        identifier = self._make_identifier(instance)

        created = kwargs.get('created')

        if klass in self._callbacks and (identifier in self._initial_data[klass] or created):
            initial_data = self._initial_data[klass].get(identifier, {})

            for key, condition, callback in self._callbacks[klass]:
                if condition(initial_data, instance, created):
                    event = self.add(key, instance)

                    if callback:
                        callback(event,
                                 initial_data=initial_data,
                                 instance=instance,
                                 created=created)

    def _on_pre_save(self, sender, instance, **kwargs):
        klass = instance.__class__

        if klass in self._callbacks:
            if not klass in self._initial_data:
                self._initial_data[klass] = {}

            if instance.pk:
                previous_instance = klass.objects.get(pk=instance.pk)

                identifier = self._make_identifier(instance)

                self._initial_data[klass][identifier] = dict(previous_instance.__dict__)

    def _make_identifier(self, instance):
        return self.make_key_id_from_instance(instance) \
            if instance.pk else id(instance)



import inspect
from datetime import datetime

from django.db.models import Model

from events_watcher.utils import load_class
from events_watcher import settings
from ..base import Backend

import redisco


class RedisBackend(Backend):
    filters = [
        (lambda instance: isinstance(instance, str), 'list_from_string'),
        (lambda instance: isinstance(instance, Model), 'list_from_instance'),
        (lambda instance: inspect.isclass(instance) and issubclass(instance, Model), 'list_from_class'),
    ]

    def __init__(self):
        super(RedisBackend, self).__init__()

        redisco.connection_setup(**settings.REDIS_CONNECTION)

        self.model = load_class(settings.REDIS_CONNECTION_CLASS)

    def add(self, key, instance, date=None):
        if not date:
            date = datetime.now()

        event = self.model(name=key,
                           reference=self.make_key_from_instance(instance),
                           date=date,
                           object_id=instance.pk)
        event.save()

        return event

    def list_from_string(self, string):
        return self.model.objects.filter(name=string)

    def list_from_instance(self, instance):
        return self.model.objects.filter(reference=self.make_key_from_instance(instance),
                                         object_id=instance.id)

    def list_from_class(self, klass):
        return self.model.objects.filter(reference=self.make_key_from_class(klass))

    def remove(self, instance):
        for event in self.list(instance):
            event.delete()

    def retrieve(self, name, instance):
        events = self.list(instance).filter(name=name)

        if len(events):
            return events[0]

        return False



from django.conf import settings


BACKEND = getattr(settings,
                  'EVENTS_WATCHER_BACKEND',
                  'events_watcher.backends.database.DatabaseBackend')

DATABASE_CONNECTION_CLASS = getattr(settings,
                                    'EVENTS_WATCHER_DATABASE_CONNECTION_CLASS',
                                    'events_watcher.backends.database.models.Event')

REDIS_CONNECTION = getattr(settings,
                           'EVENTS_WATCHER_REDIS_CONNECTION',
                           {'host': 'localhost', 'port': 6379, 'db': 0})

REDIS_CONNECTION_CLASS = getattr(settings,
                                 'EVENTS_WATCHER_REDIS_CONNECTION_CLASS',
                                 'events_watcher.backends.redis.models.Event')



from django.db import models
from django.contrib.contenttypes import generic
from django.contrib.contenttypes.models import ContentType

from ..mixins import EventMixin


class Event(models.Model, EventMixin):
    object_id = models.PositiveIntegerField()
    content_object = generic.GenericForeignKey('content_type', 'object_id')

    name = models.CharField(max_length=150)
    date = models.DateTimeField(blank=True, null=True)
    content_type = models.ForeignKey(ContentType)



from redisco import models

from ..mixins import EventMixin


class Event(EventMixin, models.Model):
    name = models.Attribute()
    object_id = models.IntegerField()
    reference = models.Attribute()
    date = models.DateTimeField(required=False)

    @property
    def content_object(self):
        return self._get_model_class().objects.get(pk=self.object_id)

    @property
    def content_type(self):
        from django.contrib.contenttypes.models import ContentType

        return ContentType.objects.get_for_model(self._get_model_class())

    def _get_model_class(self):
        from django.db import models

        return models.get_model(*self.reference.split('.'))
